"""
Welcome to GDA Score project documentation! This project is divided into 3 sub-modules. `gdaAttack` contains
functionalities related to performing an attack on an anonymized data set. `gdaScore` is used for evaluating an attack and finally
`gdaTools` contains helper and utility functions such as `setupGdaAttackParameters` which are used in other routines.

## Getting started

 - for installation, please follow the instruction at: [gdascore on PyPI](https://pypi.org/project/gda-score-code/)
 - Have a look at sample attacks here: [github repository](https://github.com/gda-score/attacks)
"""