myCredentials = '''{
  "postgres": {
    "user": "gda-score_ro_user",
    "password": "moquaiR7"
  },
  "aircloak": {
    "user": " gda-score_ro_user",
    "password": "moquaiR7"
  }
}
'''

readme = '''Create a file called `myCredentials.json` in this directory, using `example.myCredentials.json` as a template.
'''


